/***********************************************************************

 SV1Codec - Transport Tycoon savegame decompression/compression utility

 Version 0.90

 ----------------------------------------------------------------------
 Copyright (C) 2002 by Marcin Grzegorczyk

 Permission is hereby granted, free of charge, to any person
 obtaining a copy of this software and associated documentation files
 (the "Software"), to deal in the Software without restriction,
 including without limitation the rights to use, copy, modify, merge,
 publish, distribute, sublicense, and/or sell copies of the Software,
 and to permit persons to whom the Software is furnished to do so,
 subject to the following conditions:

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 ----------------------------------------------------------------------

 Transport Tycoon is Copyright (C) 1994-1995 Chris Sawyer.
 Any other trademarks that appear in the source code or documentation
 files are the property of their respective owners.

 =====================================================================

 OK, enough legalese. The point of it is, you can do with SV1Codec
 about anything you want, except claim you wrote it yourself or try
 to push it under a different license. You can recompile the source,
 make your own modifications, extract frgaments of code and put in
 your own programs, and of course you can distribute the thing around.

 Some parts of SV1Codec are in the public domain, as indicated in the
 comments at the beginning of the respective files. This means they're
 not copyrighted (I'm not going to claim copyright over things I didn't
 invent, for instance), and they're of course not subject to the above
 license, so you can reuse them freely without worrying about what is
 a "substantial portion" of the software.

 Bug fixes, feature recommendations and otherwise comments
 gladly accepted at mgrzegor@itma.pwr.wroc.pl.

 =====================================================================

 Changes from 0.10 to 0.11 (project named SV1Tool at that point):
 - more comments in the source code :-)
 - options -7, -8, -u; #define SAFE_CHARS_HIGH and all that stuff
 - landscape style autodetection code
 - TTDPatch -x signature check code
 - error reporting via strerror(), see StrError()
 - delete-on-write-error code in my_write()
 - C89_ONLY changed to C90_ONLY
   (I don't want to keep track of changes between C89 and C90)
 - accept lowercase %XX and #xXX escape sequences
 - removed free()s as they shouldn't be necessary;
   OTOH, "return EXIT_FAILURE" in main() generally replaced by
   "goto exit_failure" in case one has to free malloc()ed memory
   before exiting

 Changes from 0.11 to 0.90:
 - non-interactive tool: project renamed to SV1Codec
 - license changed to the Open Source MIT License
   (but RLE routines and related stuff are still public domain)
 - separated into several C source and header files, but in a way such
   that it's still possible to compile the whole thing by
   "$CC sv1codec.c"; uncomment the "#define SINGLE_TRANSLATION_UNIT=0"
   in sv1ccfg.h to compile the files separately and then link them
 - move away from using a lot of global variables; 'global' file-related
   parameters are now passed via structs sOpParameters and sFilesInfo,
   declared in sv1codec.h
 - new default escape sequence style: @XX
   (inspired by Win32 versions of Wget); %XX is still accepted by
   default, #xXX requires a command-line switch (-xch); also yet another
   new style: _0xXX (-xcp switch), which fulfils the minimum POSIX
   requirements for file names
 - recognize TTO playable demo savegames
   (for now, only version 1.05.116 is recognized)
 - separate info and error message streams
   (global variables 'msgout' and 'msgerr')
 - RLE_Encode() returns success/error code (write-related)
 - new feature: checksum correction; if no <output> specified on the
   command line, open <input> again for update and fwrite() there
 - *lots* of other changes...

***********************************************************************/

/*
 Note from the author:
 I decided that SV1Codec is to be written in a maximally portable way.
 Therefore, I stick to the features defined in the ANSI/ISO C Standard,
 avoiding any implementation-defined, unspecified or undefined behaviour
 as far as possible (that is, keeping the program as close to being
 strictly conforming as I can). However, some assumptions had to be made
 about the runtime environment:
 - Both the source and the execution character sets support all the
   non-control characters of the US-ASCII character set.
 - The target machine has to use 8 bits per byte, and use the two's
   complement notation for signed arithmetic on bytes. Also, 16-bit and
   32-bit integer types must be available. (See the environment
   configuration section in sv1codec.h.)
 - It is possible to create objects up to 8 MB in size, and fread() or
   fwrite() such amounts of data in one call. (If size_t is a 16-bit
   type, the program may compile and run but will refuse to deal
   with files over 64 KB in size, which will make it useless. C99
   compilers should reject the program outright in such case. Again,
   see sv1codec.h for details.)
 - The file system supports precise read/write/update/seek semantics
   for binary files; fseek() with SEEK_END as the third argument is
   well-defined.
 - Possibly more, it's really hard to keep track of all the portability
   issues the Standard mentions...
 Still, I think SV1Codec should compile out of the box, as long as the
 environment is not really weird and the compiler is ANSI C (actually,
 ISO 9899:1990) compliant. Check sv1ccfg.h for various configuration
 macros you can tweak.
*/


#include "sv1ccfg.h"        /* All the configuration macros go there */


/*
 By default, we're supposed to just compile this file, and it will
 #include all the other files. However, if SINGLE_TRANSLATION_UNIT
 is #defined as 0 (in sv1ccfg.h or on the command line), no .c file
 will include another .c file -- we must compile each of them
 separately.
*/
#ifndef SINGLE_TRANSLATION_UNIT
#define SINGLE_TRANSLATION_UNIT 1
#endif


#include "sv1codec.h"
#include "sv1cmisc.h"
#include "sv1cio.h"
#include "sv1crle.h"
#include "sv1ccli.h"
#include "sv1cui.h"


#if SINGLE_TRANSLATION_UNIT
  #include "sv1cmisc.c"
  #include "sv1cio.c"
  #include "sv1crle.c"
  #include "sv1ccli.c"
  #include "sv1cui.c"
#endif


/*
 Now we include the system stuff we need.
 Note: <stddef.h>, <limits.h>, and also <stdbool.h> and <stdint.h>
 if they're there, are included by sv1codec.h. 
*/
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>


/********** Global constants **********/

static const char landscape_letters[4] = {'T', 'C', 'H', 'Y'};

/* Program version number */
PUBLIC_DEFINITION const unsigned Version = 0x005A;

/*
 Union type used by MunchLoadedFile() to return additional parameters
 related to the file being processed
*/
union uMunchFileReturn1 {
  struct {
    uint32_t calc_chksum, file_chksum;    /* computed and stored file checksum */
  } d;                                  /* used in decode/chksum mode */
  struct {
    int ttdp;                             /* TTDPatch -x option needed, zero if not known or N/A */
    signed char ttdp_sig;                 /* TTDPatch extended array signature in the file */
  } e;                                  /* used in encode mode */
};

/********** End of globals section **********/





static
unsigned char *AllocateTitleBuffer(size_t length)
/*
 Allocate 'length'+2 or TTx_MAX_TITLE_LENGTH+2 bytes for the title buffer,
 whichever is more.
 Return pointer to the allocated buffer on success, NULL on failure.
*/
{
  assert((size_t)(length + 2) > length);
  return malloc(length > TTx_MAX_TITLE_LENGTH ? length + 2 : TTx_MAX_TITLE_LENGTH + 2);
}



static
int MunchLoadedFile(
  struct sOpParameters *const opp,          /* I/O: Operation parameters */
  struct sFilesInfo *const finfo,           /* I/O: Loaded file descriptor */
  unsigned char **const titlebufp,          /* I/O: Game title buffer */
  union uMunchFileReturn1 *const otherp)    /* Out: Other parameters, see union uMunchFileReturn1 */
/*
 After a file is loaded, do all the complicated processing to determine
 what the file type is and all the extra parameters.

 On entry, '*titlebufp' either points to the user-specified game title
 or is NULL. On successful exit, it points to the game title buffer
 (guaranteed to be at least TTx_MAX_TITLE_LENGTH+2 bytes long),
 allocated and filled in with the game title.

 Return 0 on success, nonzero on error.
*/
{
  size_t ifnl = strlen(finfo->infname);
  unsigned char *titlebuf /* DEBUG_NULLINIT(NOTHING) // not used */;

  assert(titlebufp);

  /***** Determine input file type *****/

  /*
   If -c, -d or -e is specified, the usual sv1/big detection is skipped
   by doing a goto to the appropriate location. I know that this results
   in a rather crufty code, but hey, it's the easiest way to do it :-)
  */
  switch (opp->operation) {
    case OP_DECODE:
    case OP_CHKSUM:
      /* if the savegame type is not forced (-tt[pd|o|d]), we must detect it now */
      if (!opp->sv_ver_forced && ((opp->sv_version = DetectTTByTitle(finfo->buf.ptr, finfo->buf.len)) == SV_VER_UNKNOWN)) {
        if ((opp->sv_version = UI_SaveTypeNotRecognized()) == SV_VER_UNKNOWN) return 1;
      }
      goto process_savegame;
    case OP_ENCODE: goto check_big_file_type;
    case OP_DEFAULT: ;
  }
  /* OK, back to the normal code... */

  /* If the input file has the .big extension, we don't even check its title checksum */
  if (ifnl < 4 || !StrEQ(finfo->infname + ifnl - 4, ".big")) {
    /* no .big extension, check if it looks like a savegame */
    enum eSVVersion sv_v = DetectTTByTitle(finfo->buf.ptr, finfo->buf.len);
    if (sv_v != SV_VER_UNKNOWN) {
      Bool chksum_adjusted;
      uint32_t file_chksum, calc_chksum;
      /*
       OK, so it seems to be a savegame, so let's record the type (TTO/TTD).
       But not if the type is forced on the command line; in such case,
       we assume the user knows better.
      */
      if (!opp->sv_ver_forced) opp->sv_version = sv_v;

      /*** 'sv_v' MUST NOT be used anywhere below ***/ 

      /* (LABEL) Savegame version detected or forced */
      process_savegame: finfo->buf_data_type = BUFDATA_TT_RLE;

      /* set all the format parameters */
      if (!opp->sv_form_forced) {
        opp->sv_format.checksum_adj = ChecksumAdjFromVersion(opp->sv_version);
        opp->sv_format.title_length = TitleLengthFromVersion(opp->sv_version);
        opp->sv_format.data_start = opp->sv_format.title_length + 2;
        /* Sanity check: Verify that data_length won't wrap around */
        assert(finfo->buf.len >= opp->sv_format.title_length + 8);
        /*
         The '8' here is:
         * title checksum = 2 bytes
         * minimum-length RLE sequence = 2 bytes
         * file checksum = 4 bytes
        */
        opp->sv_format.data_length = finfo->buf.len - opp->sv_format.title_length - 6;
        /* RLE data length is file length minus the title field and both checksums, hence '6' */
        UI_FileTypeMsg(opp, 0);
      }
      else {
        /* ###TODO: validate range for forced format parameters */
        /* for now, just assert false */
        #define ManualFormatNotImplemented 0
        assert(ManualFormatNotImplemented);
      }
      /* is it the TITLE.DAT file? */
      if (!opp->title_dat_forced) {
        /* not forced on the command line -- check the file name */
        opp->title_dat = ifnl >= 9 && StrEQ(finfo->infname + ifnl - 9, "TITLE.DAT");
      }
      /* if so, the checksum is to be subtracted, not added */
      if (opp->title_dat) {
        if (opp->sv_version != SV_VER_TTp_DEMO) {   /* TT Playable Demos do not invert the meaning of checksum_adj for their TITLE.DAT files */
          opp->sv_format.checksum_adj = -opp->sv_format.checksum_adj;   /* Note: we're negating an unsigned value here. This is OK. */
        }
        UI_FileTreatedAsTitleDemoMsg();
      }

      /* Allocate a buffer for the game title */
      titlebuf = AllocateTitleBuffer(opp->sv_format.title_length);
      if (titlebuf == NULL) {
        UI_OutOfMemForTitleBuf();
        return 2;
      }
      /* Copy the savegame title */
      strncpy((char *)titlebuf, (char *)finfo->buf.ptr, opp->sv_format.title_length);
      titlebuf[opp->sv_format.title_length] = '\0';
      UI_GameTitleMsg((char *)titlebuf);

      /* Read the file checksum from the last four bytes */
      assert(finfo->buf.len >= 4);
      file_chksum = GetLEInt32(finfo->buf.ptr + (finfo->buf.len - 4));
      /* Compute the actual checksum */
      UI_ComputingChecksumMsg();
      calc_chksum = CalcFileChecksum(finfo->buf.ptr, finfo->buf.len - 4, opp->sv_format.checksum_adj);
      chksum_adjusted = 0;
      if (calc_chksum != file_chksum && !opp->sv_ver_forced && opp->sv_version == SV_VER_TTO) {
        /* Check if it's a TT playable demo savegame */
        /* ###TODO: improve to support various versions */
        register uint32_t adj = opp->title_dat ? TTp_FILE_CHECKSUM + TTO_FILE_CHECKSUM : TTp_FILE_CHECKSUM - TTO_FILE_CHECKSUM;
        if ((uint32_t)(calc_chksum + adj) == file_chksum) {
          opp->sv_version = SV_VER_TTp_DEMO;
          opp->sv_format.checksum_adj = TTp_FILE_CHECKSUM;   /* Note: TT Playable Demos do not invert the meaning of checksum_adj for their TITLE.DAT files */
          calc_chksum = file_chksum;
          chksum_adjusted = 1;
        }
      }
      UI_ChecksumComputedMsg(calc_chksum, file_chksum, chksum_adjusted);
      if (chksum_adjusted) UI_IsTTPlayableDemoSaveMsg();

      if (opp->operation != OP_CHKSUM) {
        if (file_chksum != calc_chksum) {
          /* check if it might be a renamed TITLE.DAT */
          /* (N/A if parameters have been forced on the command line) */
          /* (N/A for TT Playable Demo files) */
          if (!opp->title_dat
                && !opp->title_dat_forced
                && !opp->sv_form_forced
                && opp->sv_version != SV_VER_TTp_DEMO
                && calc_chksum - opp->sv_format.checksum_adj - opp->sv_format.checksum_adj == file_chksum) {
            /* appears to be a renamed TITLE.DAT indeed */
            opp->title_dat = 1;
            opp->sv_format.checksum_adj = -opp->sv_format.checksum_adj;   /* Note: we're negating an unsigned value here. This is OK. */
            calc_chksum = file_chksum;
            UI_AppearsToBeTitleDemoMsg();
          }
          else {
            UI_FileChecksumBad();
            if (opp->ignore_errors) {
              UI_IgnoreErrorProceed();
            }
            else {
              if (!UI_Abort_CanForceProceed()) return 3;
            }
          }
        }
        assert(opp->operation == OP_DEFAULT || opp->operation == OP_DECODE);
        opp->operation = OP_DECODE;
      }
      /* Return additional parameters to the caller */
      otherp->d.calc_chksum = calc_chksum;
      otherp->d.file_chksum = file_chksum;
    }
    else {
      int ttdp;
      signed char ttdp_sig;
      unsigned char *namebuf;
      /*
       OK, this file does not seem to be a savegame.
      */

      UI_FileIsNotASavegameMsg();

      /* (LABEL) 'Big' file assumed */
      check_big_file_type: finfo->buf_data_type = BUFDATA_RAW;

      ttdp = 0;
      ttdp_sig = 0;
      /* DEBUG_NULLINIT(namebuf) // not used */;

      if (!opp->sv_ver_forced) {
        if ((opp->sv_version = DetectTTBySize((long)finfo->buf.len, &ttdp)) == SV_VER_UNKNOWN) {
          if ((opp->sv_version = UI_BigSizeNotRecognized()) == SV_VER_UNKNOWN) return 4;
        }
      }
      /* 'big' file type is now known (has been detected or forced) */
      assert(opp->sv_form_forced || opp->sv_version != SV_VER_UNKNOWN);

      /* set all the format parameters */
      if (!opp->sv_form_forced) {
        register uint32_t adj = ChecksumAdjFromVersion(opp->sv_version);
        opp->sv_format.checksum_adj = opp->title_dat && opp->sv_version != SV_VER_TTp_DEMO ? -adj : adj;   /* Note: we're negating an unsigned value here. This is OK. */
        opp->sv_format.title_length = TitleLengthFromVersion(opp->sv_version);
        opp->sv_format.data_start = 0;
        opp->sv_format.data_length = finfo->buf.len;
        UI_FileTypeMsg(opp, 1);
      }
      else {
        /* ###TODO: validate range for forced format parameters */
        /* for now, just assert false */
        #define ManualFormatNotImplemented 0
        assert(ManualFormatNotImplemented);
      }

      assert(opp->operation == OP_DEFAULT || opp->operation == OP_ENCODE);
      opp->operation = OP_ENCODE;

      /* For TTD games, check TTDPatch extended vehicle array signature */
      if (ttdp) {
        /* sanity check */
        assert(opp->sv_version == SV_VER_TTD);
        assert(!opp->sv_ver_forced);
        assert(TTD_MOREVEHICLES_OFFSET < finfo->buf.len);
        /* read the signature */
        ttdp_sig = (signed char)finfo->buf.ptr[TTD_MOREVEHICLES_OFFSET];
        if (ttdp_sig + (ttdp_sig <= 1) != ttdp) {
          UI_TTDPatchSigBad(ttdp_sig, ttdp);
          ttdp_sig = -1;
          if (opp->ignore_errors) {
            UI_IgnoreErrorProceed();
          }
          else {
            if (!UI_Abort_CanForceProceed()) return 5;
          }
        }
      }

      /* Determine name (title) of the savegame */

      if (*titlebufp != NULL) {
        /* => specified on the command line */
        namebuf = *titlebufp;
      }
      else {
        /* => must be determined from the 'big' file name */
        const char *fname, *fext;
        size_t fnl;

        /* Find the local file name (strip path) */
        fname = StripFilePath(finfo->infname);
        /* Strip the extension */
        fext = strrchr(fname, '.');
        if (fext != NULL) fnl = (size_t)(fext - fname);
        else fnl = strlen(fname);

        /* Allocate an intermediate buffer and copy the file name there */
        if ((size_t)(fnl + 1) == 0) fnl = SIZE_MAX - 1;  /* safety check */
        namebuf = malloc(fnl + 1);
        if (namebuf == NULL) {
          UI_OutOfMemForBigNameBuf();
          return 6;
        }

        strncpy((char *)namebuf, fname, fnl);
        namebuf[fnl] = '\0';
      }

      /* Allocate a buffer for the game title */
      titlebuf = AllocateTitleBuffer(opp->sv_format.title_length);
      if (titlebuf == NULL) {
        UI_OutOfMemForTitleBuf();
        return 7;
      }
      /* Put the file name or the manually specified name there, folding escape sequences */
      memset(titlebuf, 0, opp->sv_format.title_length);
      assert(namebuf);
      StrNUnfix(titlebuf, namebuf, opp->sv_format.title_length - 1, opp);
      UI_GameTitleMsg((char *)titlebuf);

      /* Deallocate the intermediate name buffer if necessary */
      if (*titlebufp == NULL) free(namebuf);

      /* Return additional parameters to the caller */
      otherp->e.ttdp = ttdp;
      otherp->e.ttdp_sig = ttdp_sig;
    }
  }
  else goto check_big_file_type;  /* .big extension */

  assert(titlebuf);
  *titlebufp = titlebuf;
  return 0;
}




static
int ConstructOutFileName(
  const struct sOpParameters *const opp,    /* In : Operation parameters */
  struct sFilesInfo *const finfo,           /* I/O: File description */
  const unsigned char *title,               /* In : Game title buffer */
  int ttdp)                                 /* In : TTDPatch -x option used (valid only in encode mode) */
/*
 If the destination is not given or is a directory (in other words,
 'finfo->outpname' is NULL or ends with a path separator), this function
 is called to determine the name the output file will have. This name
 is malloc()ed and returned in 'finfo->outfname'.

 Note: In the fix checksum mode, the lack of the destination argument
 has a special meaning (write checksum to the same file); therefore,
 this function should NOT be called if (opp->operation == OP_CHKSUM)
 and (finfo->outpname == NULL).
*/
{
  const char *const outpname = finfo->outpname != NULL ? finfo->outpname : "";
  size_t pl = strlen(outpname);

  switch (opp->operation) {
    case OP_DECODE: {
      size_t tl = strlen((char *)title);

      if (tl > SIZE_MAX / MAX_StrFix_ESCAPE_CHARS) {
        UI_OutputPathTooLong();
        return 1;
      }
      tl *= MAX_StrFix_ESCAPE_CHARS;  /* each character may expand to an escape sequence */
      if (SIZE_MAX - pl <= tl || SIZE_MAX - pl - tl < 5) {
        UI_OutputPathTooLong();
        return 2;
      }

      finfo->outfname = malloc(pl + tl + 5);  /* 4 characters for ".big" + 1 for the final NUL */
      if (finfo->outfname == NULL) {
        UI_OutOfMemForOutFileName();
        return 3;
      }
      strcpy(finfo->outfname, outpname);
      assert(strlen(finfo->outfname) == pl);  /* sanity check */
      StrFix((unsigned char *)finfo->outfname + pl, title, opp);
      strcat(finfo->outfname, ".big");
      break;
    }
    case OP_CHKSUM: {
      const char *infname = StripFilePath(finfo->infname);
      size_t fl = strlen(infname);

      assert(finfo->outpname != NULL);  /* just to be sure */

      /* Check if the file is to be saved under a regular name */
      /* is -tr[tpchy] specified? */
      if (finfo->save_letter != '?') goto find_savegame;
      /* has a regular name? */
      if (fl == 9 && toupper((unsigned char)infname[0]) == 'T'
                  && toupper((unsigned char)infname[1]) == 'R'
                  && infname[2] != '?'  /* safety check */
                  && infname[5] == '.'
                  && toupper((unsigned char)infname[6]) == 'S'
                  && strchr("SVsv", infname[7])
                  && strchr("012", infname[8])) {
        finfo->save_letter = (char)toupper((unsigned char)infname[2]);
        if (!*finfo->save_ext) {
          int i;
          for (i = 0; i < 3; i++) finfo->save_ext[i] = (char)toupper((unsigned char)infname[i+6]);
        }
        goto find_savegame;
      }
      /* nope, copy the original name */

      if (SIZE_MAX - pl <= fl) {
        UI_OutputPathTooLong();
        return 4;
      }

      finfo->outfname = malloc(pl + fl + 1);
      if (finfo->outfname == NULL) {
        UI_OutOfMemForOutFileName();
        return 5;
      }
      strcpy(finfo->outfname, outpname);

      strcat(finfo->outfname, infname);
      break;
    }
    case OP_ENCODE:
      assert(!ttdp || (opp->sv_version == SV_VER_TTD && !opp->sv_ver_forced));  /* sanity check */

      /* (LABEL) */
      find_savegame:

      if (SIZE_MAX - pl < 10) {
        UI_OutputPathTooLong();
        return 6;
      }

      finfo->outfname = malloc(pl + 10);  /* 9 characters for file name, 1 for the final NUL */
      if (finfo->outfname == NULL) {
        UI_OutOfMemForOutFileName();
        return 7;
      }
      strcpy(finfo->outfname, outpname);

      if (opp->title_dat) {
        strcat(finfo->outfname, "TITLE.DAT");
      }
      else {
        /* we'll search for an existing savegame */
        if (!*finfo->save_ext) strcpy(finfo->save_ext, "SV1");  /* default extension */
        /* ###TODO: detect the number of players */ 
        if (finfo->save_letter == '?') {
          assert(finfo->buf_data_type == BUFDATA_RAW);  /* sanity check */
          if (toupper((unsigned char)finfo->save_ext[1]) == 'S') {
            /* Encoding as a predefined scenario, autodetect climate */
            int i = GetLandscapeStyle(&finfo->buf, ttdp);
            switch (i) {
              default:
                if (i < 0) {  /* not a TTD savegame, or forced by -ttd */
                  UI_CannotDetectLandscapeStyle(opp);
                }
                else {
                  UI_BadLandscapeStyle(i);
                }
                if ((i = UI_LandscapeLetterNotKnown()) < 0) return 8;
              /* acceptable values for climate byte */
              case 0:
              case 1:
              case 2:
              case 3:;
            }
            assert(i >= 0 && i < 4);  /* sanity check */
            UI_LandscapeStyleDetectedMsg(i);
            finfo->save_letter = landscape_letters[i];
          }
          else {
            /* Not a predefined scenario, use TTDPatch 1.8 rules */
            finfo->save_letter = ttdp > 1 ? (char)'P' : (char)'T';
                    /* Note: The character constants are cast to (char) to suppress
                             the silly warning about a conversion to a narrower type
                             that the Borland C compiler would display. */
          }
        }

        if (SearchGameFile(opp, finfo, (char *)title) == NULL) {
          /* Failed to assign a file name */
          if (UI_NoFreeSaveName(finfo) == NULL) return 9;
        }
      }
      break;

    default:
      #define BadOperationCode 0
      assert(BadOperationCode);
  }

  return 0;
}




static
int DecodeToFile(const struct sOpParameters *const opp, const struct sFilesInfo *const finfo)
{
  struct sStreamInfo outs;
  int rle_result, close_result;

  /* sanity checks */
  assert(finfo->buf.ptr != NULL);
  assert(finfo->buf_data_type == BUFDATA_TT_RLE);
  assert(finfo->outfname != NULL);

  outs.filename = finfo->outfname;
  if (CreateOrOpenOutputFile(1, &outs)) return 1;

  /* Now decode to the file */
  rle_result = RLEDecodeToStream(finfo->buf.ptr + opp->sv_format.data_start, opp->sv_format.data_length, &outs);
  if (rle_result > 0) {
    /* fatal error, the end of the input buffer was not reached */
    /* the file should have been closed and deleted */
    return 2;
  }

  /*
   close the file; display "done" or "error" on msgout as appropriate

   if an I/O error occurs, don't display the full message yet;
   tell UI_FileCloseFailed() to save the error code for later
  */
  close_result = CloseOutputFile(&outs, 1);
  if (!close_result) UI_FileWrittenAndClosedMsg();

  /* display the RLE bad data warning if necessary */
  if (rle_result < 0) UI_RLEDecodeErrorWarn();
  /* now display the I/O error message if necessary, using the saved error code */
  if (close_result) {
    UI_FileCloseFailed(0, 0, 1);
    return 3;
  }

  return 0;
}



static
int EncodeToFile(
  const struct sOpParameters *const opp,
  const struct sFilesInfo *const finfo,
  unsigned char *const titlebuf)
{
  struct sStreamInfo outs;
  int close_result;
  uint16_t title_checksum;
  uint32_t file_checksum;
  unsigned char file_chksum_buffer[4];
  size_t b;

  /* sanity checks */
  assert(finfo->buf.ptr != NULL);
  assert(finfo->buf_data_type == BUFDATA_RAW);
  assert(finfo->outfname != NULL);
  assert(titlebuf != NULL);

  outs.filename = finfo->outfname;
  if (CreateOrOpenOutputFile(1, &outs)) return 1;

  /* calculate title checksum, put it after the title field */
  title_checksum = CalcTitleChecksum(titlebuf, opp->sv_format.title_length);
  SetLEInt16(titlebuf + opp->sv_format.title_length, title_checksum);

  /* write the title buffer at the beginning of the file */
  if (WriteData(titlebuf, opp->sv_format.title_length + 2, &outs)) return 2;

  /* process the title buffer for the file checksum */
  file_checksum = 0;
  for (b = 0; b < opp->sv_format.title_length + 2; b++) ChksumStep(titlebuf[b], &file_checksum);

  /* Now encode to the file */
  if (RLEEncodeToStream(finfo->buf.ptr + opp->sv_format.data_start,
                        opp->sv_format.data_length,
                        &outs,
                        &file_checksum,
                        opp->rle_code_min,
                        opp->rle_code_max) != 0) return 3;  /* write error, file should have been closed and deleted */

  /* Wrap up the file checksum calculation */
  file_checksum += opp->sv_format.checksum_adj;

  /* write the checksum at the end of the file */
  SetLEInt32(&file_chksum_buffer, file_checksum);
  if (WriteData(&file_chksum_buffer, 4, &outs)) return 4;

  /*
   close the file; display "done" or "error" on msgout as appropriate

   if an I/O error occurs, don't display the full message yet;
   tell UI_FileCloseFailed() to save the error code for later
  */
  close_result = CloseOutputFile(&outs, 1);
  if (!close_result) UI_FileWrittenAndClosedMsg();

  /* display the computed checksum */
  UI_ChecksumAfterEncodeMsg(file_checksum);
  /* now display the I/O error message if necessary, using the saved error code */
  if (close_result) {
    UI_FileCloseFailed(0, 0, 1);
    return 5;
  }

  return 0;
}



static
int SaveOutFile(const struct sFilesInfo *const finfo)
/*
 Note: function intended for copying a savegame to a different path;
 in this case, the file and title checksums are assumed to be verified
*/
{
  struct sStreamInfo outs;

  /* sanity checks */
  assert(finfo->buf.ptr != NULL);
  assert(finfo->outfname != NULL);

  outs.filename = finfo->outfname;
  if (CreateOrOpenOutputFile(1, &outs)) return 1;

  if (WriteData(finfo->buf.ptr, finfo->buf.len, &outs)) return 2;

  if (CloseOutputFile(&outs, 0)) return 3;

  UI_FileWrittenAndClosedMsg();
  return 0;
}



static
int UpdateFileChecksums(const struct sFilesInfo *const finfo, Bool update_title_chksum, size_t title_length)
/*
 Note: this function uses 'finfo->infname', since it is to be called
 when 'finfo->outfname' is NULL.

 Note: the file checksum is placed in the last four bytes of the buffer
*/
{
  struct sStreamInfo outs;

  /* sanity checks */
  assert(finfo->buf.ptr != NULL);
  assert(finfo->buf_data_type == BUFDATA_TT_RLE);
  assert(finfo->infname != NULL);

  outs.filename = finfo->infname;
  if (CreateOrOpenOutputFile(0, &outs)) return 1;

  if (update_title_chksum) {
    if (SeekFileOffset(&outs, (long)title_length)) return 2;
    if (WriteData(finfo->buf.ptr + title_length, 2, &outs)) return 3;
  }

  if (SeekFileOffset(&outs, (long)(finfo->buf.len - 4))) return 4;
  if (WriteData(finfo->buf.ptr + (finfo->buf.len - 4), 4, &outs)) return 5;

  if (CloseOutputFile(&outs, 0)) return 6;

  UI_FileWrittenAndClosedMsg();
  return 0;
}





int main(int argc, char **argv) {
  struct sCL_ProgParams ppar = {
    /* .op */                    {
      /* .sv_version         */    SV_VER_UNKNOWN,
      /* .sv_ver_forced      */    0,
      /* .sv_format          */    { 0, 0, 0, 0 },
      /* .sv_form_forced     */    0,
      /* .title_dat          */    0,
      /* .title_dat_forced   */    0,
      /* .operation          */    OP_DEFAULT,
      /* .ignore_errors      */    0,
      /* .use_8bit_filenames */    SAFE_CHARS_HIGH,
      /* .read_escape_modes  */    0,  /* will be filled in later */
      /* .write_escape_mode  */    FNAME_ESCAPE_AT,
      /* .rle_code_min       */    -124,
      /* .rle_code_max       */    124,
                                 },
    /* .fi */                    {
      /* .buf                */    { NULL, 0 },
      /* .infname            */    NULL,
      /* .outpname           */    NULL,
      /* .outfname           */    NULL,
      /* .buf_data_type      */    BUFDATA_UNKNOWN,
      /* .save_ext           */    "\0\0\0\0",
      /* .save_letter        */    '?',
                                 },
    /* .game_title          */   NULL,
    /* .help_requested      */   0,
    /* .msg_mode            */   MSG_MODE_STDMSG,
                               };
  union uMunchFileReturn1 add_par1;
  unsigned char *titlebuf = NULL;
  Bool update_existing_file = 0;
  Bool success = 0;

  /*
   Note: All pointers to the dynamically-allocated buffers must be NULL
   if the associated buffer is not allocated for "goto exit_failure"
   to work.

   Currently, such pointers are 'ppar.fi.buf.ptr', 'ppar.fi.outfname'
   and 'titlebuf'.

   Exception: if (ppar.fi.outfname == ppar.fi.outpname) it is recognized
   that 'ppar.fi.outfname' does not point to a malloc()ed buffer.
  */

  if (UI__Initialize()) return EXIT_FAILURE;

  if (argc > 1 && CL_ParseCmdLine(&ppar, argv + 1)) {
    UI_TryNoArgsForHelp();
    return EXIT_FAILURE;
  }

  UI__SetMessageMode(ppar.msg_mode);

  if (argc <= 1 || ppar.help_requested) {
    UI_ProgramGreeting();
    UI_ProgramHelp(ppar.help_requested);
    return EXIT_SUCCESS;
  }

  if (ppar.fi.infname == NULL) {
    UI_NoInputFile();
    UI_TryNoArgsForHelp();
    return EXIT_FAILURE;
  }

  if (ppar.op.read_escape_modes == 0) ppar.op.read_escape_modes = FNAME_ESCAPE_AT | FNAME_ESCAPE_URL;

  /* other command line sanity checks may be in order here... */

  if (LoadInFile(&ppar.fi)) goto exit_failure;

  titlebuf = (unsigned char *)ppar.game_title;  /* it's safe, MunchLoadedFile() won't modify the string... it may just strcpy() it elsewhere */ 
  if (MunchLoadedFile(&ppar.op, &ppar.fi, &titlebuf, &add_par1)) {
    titlebuf = NULL;
    goto exit_failure;
  }

  if (ppar.op.operation == OP_CHKSUM && ppar.fi.outpname == NULL) {
    update_existing_file = 1;
  }
  else if (ppar.fi.outpname == NULL || (*ppar.fi.outpname && strchr(g_path_separators, strchr(ppar.fi.outpname, '\0')[-1]))) {
    /* 'ppar.fi.outpname' is NULL or the last character is a path separator... */
    if (ConstructOutFileName(&ppar.op, &ppar.fi, titlebuf, ppar.fi.buf_data_type == BUFDATA_RAW ? add_par1.e.ttdp : 0)) {
      goto exit_failure;
    }
  }
  else ppar.fi.outfname = (char *)ppar.fi.outpname;  /* 'outpname' is const-qualified... be careful not to modify it! */

  switch (ppar.op.operation) {
    case OP_DECODE:
      if (DecodeToFile(&ppar.op, &ppar.fi)) goto exit_failure;
      break;
    case OP_ENCODE:
      if (EncodeToFile(&ppar.op, &ppar.fi, titlebuf)) goto exit_failure;
      if (add_par1.e.ttdp_sig > 0) {
        UI_GameRequiresMoreVehMsg(add_par1.e.ttdp);
      }
      break;
    case OP_CHKSUM:
      assert(!ppar.op.sv_form_forced);
      assert(ppar.op.sv_version != SV_VER_UNKNOWN);
      assert(ppar.fi.buf.len >= ppar.op.sv_format.title_length + 8);
      {
        uint16_t file_title_chksum = GetLEInt16(ppar.fi.buf.ptr + ppar.op.sv_format.title_length);
        uint16_t calc_title_chksum = CalcTitleChecksum(ppar.fi.buf.ptr, ppar.op.sv_format.title_length);
        /* fix the title checksum in case it's wrong */
        if (file_title_chksum != calc_title_chksum) {
          if (!UI_TitleChksumBad_AskFix(calc_title_chksum, file_title_chksum)) goto exit_failure;
          SetLEInt16(ppar.fi.buf.ptr + ppar.op.sv_format.title_length, calc_title_chksum);
          /* recalculate the file checksum */
          UI_TitleChksumFixedMsg();
          add_par1.d.calc_chksum = CalcFileChecksum(ppar.fi.buf.ptr, ppar.fi.buf.len - 4, ppar.op.sv_format.checksum_adj);
          UI_FileChksumRecomputedMsg();
        }
        /* fix the file checksum as well */
        if (add_par1.d.file_chksum != add_par1.d.calc_chksum) {
          SetLEInt32(ppar.fi.buf.ptr + (ppar.fi.buf.len - 4), add_par1.d.calc_chksum);
          UI_FileChksumFixedMsg();
        }

        if (update_existing_file) {
          if (add_par1.d.file_chksum == add_par1.d.calc_chksum && file_title_chksum == calc_title_chksum) {
            UI_FileChksumOKNoActionMsg();
          }
          else {
            if (UpdateFileChecksums(&ppar.fi, file_title_chksum != calc_title_chksum, ppar.op.sv_format.title_length)) goto exit_failure;
          }
        }
        else {
          if (SaveOutFile(&ppar.fi)) goto exit_failure;
        }
      }
      break;
    default:
      #define BadOperationCode 0
      assert(BadOperationCode);
  }

  success = 1;

  exit_failure:   /* if jumped to from the above, 'success' will be still 0 */

  /* Free all the dynamically-allocated buffers */
  if (ppar.fi.buf.ptr != NULL) free(ppar.fi.buf.ptr);
  if (ppar.fi.outfname != ppar.fi.outpname && ppar.fi.outfname != NULL) free (ppar.fi.outfname);
  if (titlebuf != NULL) free(titlebuf);

  return success ? EXIT_SUCCESS : EXIT_FAILURE;
}

