/***********************************************************************

 SV1Codec main header file
 (with preprocessor-based automatic environment configuration)

 This file is a part of SV1Codec.
 SV1Codec is Copyright (C) 2002 by Marcin Grzegorczyk.
 See sv1codec.c for details.

 The code in this file is explicitly placed in the public domain.
 Use, abuse and redistribute freely.

***********************************************************************/

#ifndef SV1CODEC_H
#define SV1CODEC_H


#include "sv1ccfg.h"


/*
 Let's detect the compiler used, so that we can support some of the
 extra features offered.

 For now, we only try to use a feature to suppress warnings about
 unused function arguments, which are quite common in sv1ccli.c.

 NOTE: to be pedantic, this autodetection makes the code not strictly
 conforming (as per the Standard), so we offer the macro NO_DETECT_CC,
 which, if defined, will make the compiler skip the entire
 autodetection block, thus restoring the strict ANSI/ISO conformance.
 This macro may be defined on the command line, or the definition
 may be uncommented in sv1ccfg.h.
*/
#ifndef NO_DETECT_CC
  #ifdef __BORLANDC__
    /* BCC quirks */
    #define HAVE_PRAGMA_ARGSUSED
    /* other? */
  #endif
  #ifdef __GNUC__
    /* gcc extras */
    #define PARAMETER_UNUSED(p) p __attribute__ ((__unused__))
    /* other? */
  #endif
#endif

/* these must be defined */
#ifndef PARAMETER_UNUSED
#define PARAMETER_UNUSED(p) p   /* an identity definition that won't affect anything */ 
#endif



/*
 As the code is supposed to be portable, we'll check the runtime
 environment now, using the standard macros.
*/
#include <limits.h>

/*
 To simplify the code, we rule out weird environments
 (such as 9-bit-per-byte or 1's-complement machines).
*/
#if (CHAR_BIT != 8) || (SCHAR_MIN != -128)
  #error This program requires an 8-bit 2s-complement environment.
  /* BTW, we can't use the apostrophe in an #error line! */
#endif

/*
 Check if the compiler claims to be C99-compliant.

 Some pre-C95 compilers may predefine __STDC_VERSION__ to be something
 weird, like a string constant. For this reason we have the ASSUME_C90
 macro, which, if defined, prevents the STDC version detection.
*/
#ifndef ASSUME_C90
  #if __STDC_VERSION__ >= 199901L
    /*
     If the compiler is C99-compliant, it must have <stdint.h>
     and <stdbool.h>
    */
    #ifndef HAVE_STDINT_H
      #define HAVE_STDINT_H
    #endif
    #ifndef HAVE_STDBOOL_H
      #define HAVE_STDBOOL_H
    #endif
  #endif
#endif

/* We need exact 16- and 32-bit types for checksum calculations. */
#ifdef HAVE_STDINT_H
  #ifndef NAME_STDINT_H
    #define NAME_STDINT_H <stdint.h>
  #endif
  #include NAME_STDINT_H
  #ifndef UINT16_MAX
    #error A 16-bit data type is required to compile. Sorry.
  #endif
  #ifndef UINT32_MAX
    #error A 32-bit data type is required to compile. Sorry.
  #endif
  #if SIZE_MAX < 8388607uL
    #error Compiler does not handle large enough objects. Sorry.
  #endif
#endif
#ifdef HAVE_STDBOOL_H
  #ifndef NAME_STDBOOL_H
    #define NAME_STDBOOL_H <stdbool.h>
  #endif
  #include NAME_STDBOOL_H
  #ifndef Bool
    #define Bool bool
  #endif
#endif

#ifndef SIZE_MAX
  /* no <stdint.h>, we've got to determine these types ourselves (from limits.h) */
  #if USHRT_MAX != 65535u
    #error A 16-bit data type is required to compile. Sorry.
  #endif
  typedef unsigned short uint16_t;
  #if ULONG_MAX == 4294967295uL
    typedef unsigned long uint32_t;
  #elif UINT_MAX == 4294967295uL
    typedef unsigned int uint32_t;
  #else
    #error A 32-bit data type is required to compile. Sorry.
  #endif
  /*
   Before C99, there's no standard way to determine sizeof(size_t)
   at compile time. Therefore, we must rely on runtime checks.
  */
  #define SIZE_MAX ((size_t)-1)
#endif
#ifndef Bool
  /*
   No actual boolean type exists, so we'll use 'int' to store boolean
   values. We must be careful when converting to 'Bool' from other types,
   though; use (x != 0) whenever the expression is not boolean.
  */
  typedef int Bool;
#endif




/*
 If we compile everything as a single translation unit, the 'public'
 declarations and definitions do not have to be actually public
 (i.e., they don't have to have external linkage). If we split
 the program into several units (the .c files), those declarations
 and definitions must have external linkage.

 WARNING: don't use PUBLIC_DECLARATION to declare an identifier
 that has an incomplete type (such as an array of unspecified size).
 The Standard specifically forbids tentative definitions of
 incompletely typed identifiers with internal linkage. (A tentative
 definition is a file-scope declaration with no initializer and either
 no storage class specifier or the specifier 'static'.) Identifiers
 with an incomplete type must either be declared only once (with an
 explicit initializer) if SINGLE_TRANSLATION_UNIT is true, or must have
 external linkage regardless of the value of SINGLE_TRANSLATION_UNIT.
*/
#if SINGLE_TRANSLATION_UNIT
  #define PUBLIC_DECLARATION static
  #define PUBLIC_DEFINITION static
#else
  #define PUBLIC_DECLARATION extern
  #define PUBLIC_DEFINITION
#endif




#include <stddef.h>  /* we'll need it, too */


/********** Public global identifiers **********/

PUBLIC_DECLARATION const unsigned Version;  /* program version number */


/********** Public types and defines **********/

enum eSVVersion { SV_VER_UNKNOWN, SV_VER_TTp_DEMO, SV_VER_TTO, SV_VER_TTD };
enum eMainOp { OP_DEFAULT, OP_DECODE, OP_ENCODE, OP_CHKSUM };
enum eBufDataType { BUFDATA_UNKNOWN, BUFDATA_RAW, BUFDATA_TT_RLE };

/* Modes of escape sequence prefixes in file names */
#define FNAME_ESCAPE_AT     0x01    /* @ -- default */
#define FNAME_ESCAPE_URL    0x02    /* % -- URL-style */
#define FNAME_ESCAPE_HASH   0x04    /* #x -- backward compatibility, deprecated */
#define FNAME_ESCAPE_POSIX  0x08    /* _0x -- for strict POSIX environments */


/*
 SV? format description structure.
 Note: for encoding, 'data_start' and 'data_length' refer to the data
 loaded from the 'big' file (the default is the entire file),
 not the resulting SV? file. 
*/
struct sSVFormat {
  size_t title_length;                  /* Length of the savegame title field (39 for TTO, 47 for TTD) */
  size_t data_start;                    /* Offset to the encoded data, default is 'title_length+2' (but see the note above) */
  size_t data_length;                   /* Length of the encoded data, default is the rest of the file except the last 4 bytes (see also the note above) */
  uint32_t checksum_adj;                /* The value added at the end of the checksum computing (indicates the exact savegame version) */
};


/*
 All details of the operation being performed.
*/
struct sOpParameters {
  enum eSVVersion sv_version;           /* Version of the savegame format */
  Bool sv_ver_forced;                   /* Version forced on the command line */
  struct sSVFormat sv_format;           /* Exact format of the game */
  Bool sv_form_forced;                  /* Exact format forced on the command line */
  Bool title_dat;                       /* Is the TITLE.DAT file */
  Bool title_dat_forced;                /* TITLE.DAT forced on the command line */
  enum eMainOp operation;               /* Operation to perform */
  Bool ignore_errors;                   /* -f option indicator */
  Bool use_8bit_filenames;              /* Use IS_CHAR_HIGH_SAFE() macro to determine if a character is filesystem-safe */
  unsigned read_escape_modes;           /* Which escape sequences are recognized (a combination of FNAME_ESCAPE_* defines) */
  unsigned write_escape_mode;           /* Which escape sequences are generated (one of FNAME_ESCAPE_* defines) */
  int rle_code_min;                     /* Lowest RLE code used by the encoding routine, range -128..-1 */
  int rle_code_max;                     /* Highest RLE code used by the encoding routine, range 1..128 */
};
/*
 Nb.: if the format is forced, 'sv_ver_forced' shall be also true,
 and 'sv_version' shall equal 'SV_VER_UNKNOWN'.
*/


/*
 Generic data buffer description structure.
*/
struct sBufferInfo {
  unsigned char *ptr;
  size_t len;
};


/*
 The entire state of the file being worked on.
*/
struct sFilesInfo {
  struct sBufferInfo buf;               /* The file buffer */
  const char *infname;                  /* Input file name */
  const char *outpname;                 /* Output file or directory name (may be NULL) */
  char *outfname;                       /* Output file name, as passed to a file open function */
  enum eBufDataType buf_data_type;      /* Type of data in the buffer */
  char save_ext[4];                     /* Extension to encode the savegame with */
  char save_letter;                     /* Third letter of the savegame file to use */
};


/********** Useful macros **********/

/* The size of a smallest imaginable file with the SV? file format */
#define SVx_MIN_FILE_SIZE       8           /* 1+2+1+4 */

#define TTO_TITLE_LENGTH        39
#define TTD_TITLE_LENGTH        47
#define TTx_MAX_TITLE_LENGTH    47
#define TTO_FILE_CHECKSUM       105128uL
#define TTp_FILE_CHECKSUM       105116uL    /* ###TODO: 105126 also spotted */
#define TTD_FILE_CHECKSUM       201100uL
#define TTO_BIG_SIZE            431901L
#define TTp_BIG_SIZE            430894L     /* ###TODO: 431871 also spotted */
#define TTD_BIG_SIZE            618873L
/* ###TODO: make a table of savegame version vs. big_size, perhaps offsets as well */
#define TTD_BIG_XSIZE           108800L     /* TTDPatch vehicle array extensions */
#define TTD_VEHICLEARRAY_OFFSET 0x547F2
#define TTD_CLIMATE_OFFSET      0x77131
#define TTD_MOREVEHICLES_OFFSET 0x24CBA
#define TTD_MAX_MOREVEHICLES    40

/*
 Macro to compute an offset of something in a TTD savegame, based on
 the previously determined 'ttdp' vehicle array extension of the file.
 'ttdp' is number of vehicle array entries divided by 850 (the meaning
 of the 'morevehicles' switch in TTDPatch). 'x' is the offset in a
 standard (not TTDPatch extended) savegame. 
 Note: 'ttdp' is supposed to be based on the uncompressed file size,
 not the actual TTDPatch signature at TTD_MOREVEHICLES_OFFSET.
*/
#define TTDOffset(x, ttdp) ((x) >= TTD_VEHICLEARRAY_OFFSET+TTD_BIG_XSIZE ? (x)+TTD_BIG_XSIZE*((ttdp)-1) : (x))

/* Characters defined by POSIX as safe to use in file names */
/* spacing underscore ('_') excluded as it starts an escape sequence prefix */ 
#define POSIX_SAFE_CHARS "-0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"


/*
 Macro to initialize things to zero (or NULL, in case of pointers)
 for debugging purposes. May be used in statements, as

   DEBUG_NULLINIT(foo);

 or in initialization, as

   char *foo = DEBUG_NULLINIT(NOTHING);

 The NOTHING macro (below) is needed so that you can call
 DEBUG_NULLINIT() with an empty argument; a call with no arguments
 would result in a syntax error.
*/
#ifdef NDEBUG
  #define DEBUG_NULLINIT(i)
#else
  #define DEBUG_NULLINIT(i) i = 0
#endif
#define NOTHING


/* more useful macros to come, possibly... */

/********** End of the useful macros section **********/


#endif  /* SV1CODEC_H */

