/***********************************************************************

 SV1Codec user interface routines

 This file is a part of SV1Codec.
 SV1Codec is Copyright (C) 2002 by Marcin Grzegorczyk.

 Permission is granted to copy and redistribute this file under the
 terms of the Open Source MIT License. For more information please refer
 to the file sv1codec.c or to the SV1Codec documentation.

***********************************************************************/

#include "sv1cui.h"

#include <stdio.h>
#include <string.h>
#include <assert.h>


/* The informational and error message streams (normally, stdout and stderr) */
static FILE *msgout, *msgerr;

static const char *const landscape_styles[4] = {"temperate", "sub-arctic", "sub-tropical", "toyland"};



/*
 Note the convention: functions that write informational output
 have names ending with "Msg".

 The 'err' parameter is expected to have a value taken from the 'errno'
 (the standard global identifier, declared in <errno.h>).
*/


#define StrError_MAXLEN 64  /* Maximum length of a string returned by strerror() */
static
char *StrError(int err)
/*
 Call strerror() and snip the resulting string at the first newline
 character, if any.
*/
{
  static char errbuf[StrError_MAXLEN + 1];
  char *p;

  strncpy(errbuf, strerror(err), StrError_MAXLEN);
  if ((p = strchr(errbuf, '\n')) != NULL) *p = '\0';
  return errbuf;
}



static
void UI_OpErrorMsg(void)
/*
 Called to indicate that a possibly long operation (before which the
 msgout output ended with "...") resulted in an error; a msgerr output
 will follow
*/
{
  fprintf(msgout, " Error!\n");
  fflush(msgout);  /* prepare for the following msgerr output */
}



static
void UI_OpDoneMsg(void)
/*
 Called to indicate that a possibly long operation (before which the
 msgout output ended with "...") resulted in an success
*/
{
  fprintf(msgout, " done.\n");
}



static
void UI_NLMsg(void)
/*
 Send a newline to msgout, flush msgout
*/
{
  fprintf(msgout, "\n");
  fflush(msgout);
}



#ifdef HAVE_PRAGMA_ARGSUSED
#pragma argsused
#endif
PUBLIC_DEFINITION
void UI_WriteError(PARAMETER_UNUSED(const char *filename), const int err)
/*
 Report a write error

 'filename' is currently unused
 (Note: it may be NULL (if writing to stdout))
*/
{
  UI_OpErrorMsg();
  fprintf(msgerr, "** Write error (%s).\n", StrError(err));
}



#ifdef HAVE_PRAGMA_ARGSUSED
#pragma argsused
#endif
PUBLIC_DEFINITION
void UI_SeekError(PARAMETER_UNUSED(const char *filename), const int err)
/*
 Report a seek error

 'filename' is currently unused
 (Note: it may be NULL (if writing to stdout))
*/
{
  UI_OpErrorMsg();
  fprintf(msgerr, "** File seek error (%s).\n", StrError(err));
}



PUBLIC_DEFINITION
void UI_IOFail_DeletingFile(const char *filename)
/*
 Called when a write operation failed and the file is being deleted
 (see WriteData() and FileIOFail() in sv1cio.c)
*/
{
  fprintf(msgerr, "   Deleting \"%s\"...", filename);
  fflush(msgerr);
}



PUBLIC_DEFINITION
void UI_IOFail_DeleteFileResult(const Bool failed, const int err)
/*
 Called after the attempt to delete the file
*/
{
  if (failed) fprintf(msgerr, "\n** Cannot delete the file (%s)\n", StrError(err));
  else fprintf(msgerr, " done.\n");
}



#ifdef HAVE_PRAGMA_ARGSUSED
#pragma argsused
#endif
PUBLIC_DEFINITION
void UI_SearchForGameMsg(PARAMETER_UNUSED(const char *path))
/*
 Called at the beginning of SearchGameFile() in sv1cio.c

 'path' is currently unused
*/
{
  fprintf(msgout, "Searching for the savegame file... ");
  fflush(msgout);
}



#ifdef HAVE_PRAGMA_ARGSUSED
#pragma argsused
#endif
PUBLIC_DEFINITION
void UI_GameFoundMsg(PARAMETER_UNUSED(const char *path))
/*
 Called by SearchGameFile() when the search succeeds

 'path' is currently unused
*/
{
  fprintf(msgout, "Found\n");
}



PUBLIC_DEFINITION
void UI_GameNotFoundMsg(void)
/*
 Called by SearchGameFile() when the search fails
*/
{
  fprintf(msgout, "Not found\n");
}



PUBLIC_DEFINITION
void UI_InputFileMsg(const char *const filename)
/*
 Called by LoadInFile() before opening the input file
*/
{
  fprintf(msgout, "Opening \"%s\"...", filename);
  fflush(msgout);
}



PUBLIC_DEFINITION
void UI_FileOpenOrCreateFailed(const Bool create, const int err)
/*
 Called by various functions in sv1cio.c if a file open/create operation
 fails
*/
{
  UI_OpErrorMsg();
  fprintf(msgerr, "** Cannot %s the file (%s)\n", create ? "create" : "open", StrError(err));
}



PUBLIC_DEFINITION
void UI_CannotGetFileSize(const int err)
/*
 Called by AllocAndLoadOpenInFile() if file size cannot be determined
*/
{
  UI_OpErrorMsg();
  fprintf(msgerr, "** Cannot get the file size (%s)\n", StrError(err));
}



PUBLIC_DEFINITION
void UI_FileSizeMsg(const long size)
/*
 Called by AllocAndLoadOpenInFile() after the file has been opened
 and its size determined
*/
{
  UI_OpDoneMsg();
  fprintf(msgout, "Size: %ld byte(s).", size);
}



PUBLIC_DEFINITION
void UI_FileTooBig(void)
/*
 Called by AllocAndLoadOpenInFile() if file size doesn't fit in size_t
*/
{
  UI_NLMsg();
  fprintf(msgerr, "** File is too large to load!\n");
}



PUBLIC_DEFINITION
void UI_FileTooSmall(void)
/*
 Called by AllocAndLoadOpenInFile() if file size is less than SVx_MIN_FILE_SIZE       
*/
{
  UI_NLMsg();
  fprintf(msgerr, "File is too small!\n");
}



static
void UI_OutOfMemory(void)
/*
 Generic out-of-memory message
*/
{
  fflush(msgout);
  fprintf(msgerr, "** Out of memory!\n");
}



PUBLIC_DEFINITION
void UI_OutOfMemForFileBuf(void)
/*
 Called by AllocAndLoadOpenInFile() if malloc fails
*/
{
  UI_NLMsg();
  UI_OutOfMemory();
}



PUBLIC_DEFINITION
void UI_LoadingMsg(void)
/*
 Called before loading a file (the name of which has been passed as
 an argument to UI_InputFileMsg() on a preceding call)
*/
{
  fprintf(msgout, "  Loading...");
  fflush(msgout);
}



PUBLIC_DEFINITION
void UI_ErrorReadingFile(const int err)
/*
 Called when a file load operation fails
*/
{
  UI_OpErrorMsg();
  fprintf(msgerr, "** Read error (%s)\n", StrError(err));
}



PUBLIC_DEFINITION
void (*const UI_LoadedMsg)(void) = UI_OpDoneMsg;
/*
 Called when a file load operation succeeds
*/



static
const char *UI_TTTypeSubStr(const enum eSVVersion sv_version)
{
  switch (sv_version) {
    default:
      #define TTTypeBad 0
      assert(TTTypeBad);
    case SV_VER_UNKNOWN:
      #define TTTypeMustBeKnown 0
      assert(TTTypeMustBeKnown);
    case SV_VER_TTO:      return "";
    case SV_VER_TTp_DEMO: return " Playable Demonstration";
    case SV_VER_TTD:      return " Deluxe";
  }
}



PUBLIC_DEFINITION
void UI_FileTypeMsg(const struct sOpParameters *const opp, const Bool big)
/*
 Called by MunchLoadedFile() in sv1codec.c after the initial file type
 recognition logic
*/
{
  fprintf(msgout, "File %s as a Transport Tycoon%s %s.\n",
          opp->sv_ver_forced ? "will be treated" : "recognized",
          UI_TTTypeSubStr(opp->sv_version),
          big ? "'big' file" : "savegame");
}



PUBLIC_DEFINITION
enum eSVVersion UI_SaveTypeNotRecognized(void)
/*
 Called by MunchLoadedFile() when the savegame type recognition logic
 fails

 [Reserved for an interactive version:]
 Return a manually selected savegame type
*/
{
  fflush(msgout);
  fprintf(msgerr,
          "Savegame type not recognized!\n%s",
          "Aborting. Use -ttpd, -tto or -ttd to specify game version.\n");
  return SV_VER_UNKNOWN;
}



PUBLIC_DEFINITION
void (*const UI_OutOfMemForTitleBuf)(void) = UI_OutOfMemory;
/*
 Called by MunchLoadedFile() when the title buffer cannot be allocated
*/



PUBLIC_DEFINITION
void UI_FileTreatedAsTitleDemoMsg(void)
/*
 Called by MunchLoadedFile() to report that the file will be treated
 as a TITLE.DAT (because that's its name, or the user requested it
 via the -td switch)
*/
{
  fprintf(msgout, "File will be treated as the %s (%s).\n", "main menu background savegame", "TITLE.DAT");
}



PUBLIC_DEFINITION
void UI_GameTitleMsg(const char *const title)
/*
 Called by MunchLoadedFile() to display the savegame title
 (the one that appears in the savegame selection box in TT)
*/
{
  fprintf(msgout, "Game title: %s", title);
  fprintf(msgout, "\n");  /* necessary here since the title may contain a soft EOF */
}



PUBLIC_DEFINITION
void UI_ComputingChecksumMsg(void)
/*
 Called by MunchLoadedFile() before computing the file checksum
*/
{
  fprintf(msgout, "Computing file checksum...");
  fflush(msgout);
}



PUBLIC_DEFINITION
void UI_ChecksumComputedMsg(const uint32_t cc, const uint32_t fc, const Bool adj)
/*
 Called by MunchLoadedFile() after computing the file checksum
 'cc' is the computed checksum
 'fc' is the checksum stored in the file
 'adj' indicates that 'cc' has been adjusted after the computation
 (because the file has been recognized as a TT Playable Demo savegame)
*/
{
  fprintf(msgout, " done.  Checksum: 0x%.8lX (", (unsigned long)cc);
  if (cc != fc) fprintf(msgout, "In file: 0x%.8lX", (unsigned long)fc);
  else if (adj) fprintf(msgout, "adjusted");
  else fprintf(msgout, "OK");
  fprintf(msgout, ")\n");
}



PUBLIC_DEFINITION
void UI_IsTTPlayableDemoSaveMsg(void)
/*
 Called by MunchLoadedFile() if the file is recognized
 as a TT Playable Demo savegame
*/
{
  fprintf(msgout, "This is a Transport Tycoon%s savegame.\n", UI_TTTypeSubStr(SV_VER_TTp_DEMO));
}



PUBLIC_DEFINITION
void UI_AppearsToBeTitleDemoMsg(void)
/*
 Called by MunchLoadedFile() if the file seems to be a renamed TITLE.DAT
*/
{
  fprintf(msgout, "This appears to be a renamed %s (%s).\n", "main menu background savegame", "TITLE.DAT");
}



PUBLIC_DEFINITION
void UI_FileChecksumBad(void)
/*
 Called by MunchLoadedFile() if the savegame's file checksum is incorrect

 Followed by a call to UI_IgnoreErrorProceed() or UI_Abort_CanForceProceed()
*/
{
  fflush(msgout);
  fprintf(msgerr, "File checksum is incorrect!");
}



PUBLIC_DEFINITION
void UI_IgnoreErrorProceed(void)
/*
 Called on a nonfatal error if -f was specified
*/
{
  /* always called just after another msgerr output */
  fprintf(msgerr, "  (-f option specified, proceeding)\n");
}



PUBLIC_DEFINITION
Bool UI_Abort_CanForceProceed(void)
/*
 Called on a nonfatal error if -f was not specified

 [Reserved for interactive versions:]
 Return true to ignore the error
*/
{
  /* always called just after another msgerr output */
  fprintf(msgerr, "\nAborting. Use -f to proceed anyway.\n");
  return 0;
}



PUBLIC_DEFINITION
void UI_FileIsNotASavegameMsg(void)
/*
 Called by MunchLoadedFile() if the file does not have a 'big' extension,
 but does not seem to be a savegame either
*/
{
  fprintf(msgout, "File is not a Transport Tycoon or Transport Tycoon Deluxe savegame.\n");
}



PUBLIC_DEFINITION
enum eSVVersion UI_BigSizeNotRecognized(void)
/*
 Called by MunchLoadedFile() if a call to DetectTTBySize()
 returns SV_VER_UNKNOWN

 [Reserved for interactive versions:]
 Return a manually selected file type
*/
{
  fflush(msgout);
  fprintf(msgerr,
          "File size does not match any of the known 'big' file sizes.\n%s",
          "Aborting. Use -ttpd, -tto or -ttd to specify game version.\n");
  return SV_VER_UNKNOWN;
}



PUBLIC_DEFINITION
void UI_TTDPatchSigBad(const signed char ttdp_sig, const int ttdp)
/*
 Called by MunchLoadedFile() if the TTDPatch morevehicles byte
 does not match the 'big' file size

 Followed by a call to UI_IgnoreErrorProceed() or UI_Abort_CanForceProceed()
*/
{
  fflush(msgout);
  fprintf(msgerr, "Incorrect TTDPatch vehicle array extension byte!\n"
                  "  (is %d, file size indicates %d).", ttdp_sig, ttdp);
}



PUBLIC_DEFINITION
void (*const UI_OutOfMemForBigNameBuf)(void) = UI_OutOfMemory;
/*
 Called by MunchLoadedFile() when the 'big' file name buffer
 cannot be allocated
*/



static
void UI_ErrorPrefix(void)
{
  fflush(msgout);
  fprintf(msgerr, "Error: ");
}



PUBLIC_DEFINITION
void UI_UnrecognizedOption(const char *s, const size_t l)
/*
 Print an 'Unrecognized option' error message (what else?).
 's' points to the option in question, 'l' is the number of characters
 the option consists of.
*/
{
  UI_ErrorPrefix();
  fprintf(msgerr, "Unrecognized option: -%.*s\n", (int)l, s);
}



PUBLIC_DEFINITION
void UI_MissingOptionArgument(const char *s, const size_t l)
/*
 Print a message indicating that a command-line option is missing
 its argument. 's' points to the option in question, 'l' is the number
 of characters the option consists of.
*/
{
  UI_ErrorPrefix();
  fprintf(msgerr, "Option -%.*s requires an argument\n", (int)l, s);
}



PUBLIC_DEFINITION
void UI_TooManyFilenames(void)
{
  UI_ErrorPrefix();
  fprintf(msgerr, "Too many file names specified on the command line\n");
}



PUBLIC_DEFINITION
void UI_OutputPathTooLong(void)
/*
 Called by ConstructOutFileName() in sv1codec.c when the buffer for the
 output file+path name cannot be allocated because the length it may
 need exceeds the range of size_t
*/
{
  UI_ErrorPrefix();
  fprintf(msgerr, "Output path is too long!\n");
}



PUBLIC_DEFINITION
void (*const UI_OutOfMemForOutFileName)(void) = UI_OutOfMemory;
/*
 Called by ConstructOutFileName() when the buffer for the output
 file+path name could not be malloc'ed
*/



PUBLIC_DEFINITION
void UI_CannotDetectLandscapeStyle(const struct sOpParameters *const opp)
/*
 Called by ConstructOutFileName() when the landscape style (a.k.a. climate)
 cannot be determined as 'ttdp' (the TTDPatch vehicle array extension
 indicator) is not set 
*/
{
  fflush(msgout);
  fprintf(msgerr, "Cannot detect landscape style!  (%s)\n",
          opp->sv_version == SV_VER_TTD ?
                  "-ttd option used" : "not a Transport Tycoon Deluxe file");
}



PUBLIC_DEFINITION
void UI_BadLandscapeStyle(int style)
/*
 Called by ConstructOutFileName() when the landscape style code
 does not match any of the four codes used by TTD
*/
{
  fflush(msgout);
  fprintf(msgerr, "Unknown landscape style!  (code 0x%.2X).\n", style);
}



PUBLIC_DEFINITION
int UI_LandscapeLetterNotKnown(void)
/*
 Called when the landscape style must be known but is not;
 preceded by UI_CannotDetectLandscapeStyle() or UI_BadLandscapeStyle()

 [Reserved for interactive versions:]
 Return a manually selected landscape style (0..3), or <0 to abort
*/
{
  /* always called just after another msgerr output */
  fprintf(msgerr, "Aborting. Use -trt, -trc, -trh or -try to specify landscape style manually.\n");
  return -1;
}



PUBLIC_DEFINITION
void UI_LandscapeStyleDetectedMsg(int style)
/*
 Called by ConstructOutFileName() to report the detected landscape style
*/
{
  assert(style >= 0 && style < sizeof landscape_styles / sizeof landscape_styles[0]);
  fprintf(msgout, "Landscape style detected as '%s'.\n", landscape_styles[style]);
}



#ifdef HAVE_PRAGMA_ARGSUSED
#pragma argsused
#endif
PUBLIC_DEFINITION
char *UI_NoFreeSaveName(PARAMETER_UNUSED(struct sFilesInfo *const finfo))
/*
 Called by ConstructOutFileName() if a call to SearchGameFile() fails

 [Reserved for interactive versions:]
 Return 'finfo->outfname' filled in with a user-selected path,
 or NULL to abort

 'finfo' is currently unused
*/
{
  fflush(msgout);
  fprintf(msgerr,"** Error: Cannot create a new savegame (no available file names).\n"
                 "   Overwrite an existing file or specify a different directory.");
  return NULL;
}



PUBLIC_DEFINITION
void UI_SavingAsMsg(const char *path)
/*
 Called by CreateOrOpenOutputFile() in sv1cio.c when the file is to be
 created from scratch
*/
{
  fprintf(msgout, "Saving as \"%s\"...", path);
  fflush(msgout);
}



PUBLIC_DEFINITION
void UI_UpdatingFileMsg(const char *path)
/*
 Called by CreateOrOpenOutputFile() in sv1cio.c when the file is to be
 opened for update
*/
{
  fprintf(msgout, "Updating \"%s\"...", path);
  fflush(msgout);
}



PUBLIC_DEFINITION
void UI_FileCloseFailed(const int err, const Bool out_msgflag, const Bool err_msgflag)
/*
 Called by CloseOutputFile() in sv1cio.c and other functions when the
 file close operation fails on a file being written to.

 'out_msgflag' and 'err_msgflag' control the display of both parts of
 the message (there's the msgout and the msgerr part). If 'out_msgflag'
 is FALSE and 'err_msgflag' is TRUE, 'err' is ignored and the previous
 value is used instead. This is useful if other messages may appear
 before the msgerr part.
*/
{
  static int last_error;

  if (out_msgflag || !err_msgflag) last_error = err;
  if (out_msgflag) UI_OpErrorMsg();
  fflush(msgout);
  if (err_msgflag) fprintf(msgerr, "** Failed to close the output file (%s)\n", StrError(last_error));
}



PUBLIC_DEFINITION
void (*const UI_FileWrittenAndClosedMsg)(void) = UI_OpDoneMsg;
/*
 Called after a file has been successfully written to and closed.
*/



PUBLIC_DEFINITION
void UI_RLEDecodeErrorWarn(void)
/*
 Called by DecodeToFile() in sv1codec.c if an unexpected EOF occurred
 in the RLE decode routine
*/
{
  fflush(msgout);
  fprintf(msgerr, "Warning: Unexpected end of RLE data. Input file is corrupted!\n");
}



PUBLIC_DEFINITION
void UI_ChecksumAfterEncodeMsg(unsigned long file_checksum)
/*
 Called by EncodeToFile() in sv1codec.c to report the file checksum
 after the savegame has been encoded. 
*/
{
  fprintf(msgout, "Checksum: 0x%.8lX\n", file_checksum);
}



PUBLIC_DEFINITION
int UI__Initialize(void)
/*
 User interface initialization function; return nonzero on error
*/
{
  msgout = DEFAULT_MSGOUT;
  msgerr = DEFAULT_MSGERR;
  return 0;
}



PUBLIC_DEFINITION
void UI__SetMessageMode(enum eMessageOutputMode mode)
{
  switch (mode) {
    case MSG_MODE_STDMSG:
      msgout = stdout;
      msgerr = stderr;
      break;
    case MSG_MODE_STDOUT:
      msgout = stdout;
      msgerr = stdout;
      break;
    case MSG_MODE_STDERR:
      msgout = stderr;
      msgerr = stderr;
      break;
  }
}



PUBLIC_DEFINITION
void UI_ProgramGreeting(void)
{
  fprintf(msgout, "\nSV1Codec - Transport Tycoon [Deluxe] SV1 Format Encoder/Decoder"
                  "\nVersion %d.%.2d\n\n", Version >> 8, Version & 0xFF);
}



PUBLIC_DEFINITION
void UI_ProgramHelp(Bool full_help)
{
  fprintf(msgout, "Usage: sv1codec [options] [-i] <input-file> [[-o] <output>]\n"
                  "\n"
                  "<output> may be either <output-file> or <output-directory%c>\n"
                  "\n",
                  DEFAULT_PATH_SEPARATOR);
  if (!full_help) fprintf(msgout,
         "Most useful options:\n"
         "  -c          Correct the checksums of a savegame\n"
         "  -f          Force processing even if the input file is bad\n"
         "  -sv{0,1,2}  Specify savegame extension (default -sv1)\n"
         "  -ss{0,1}    Use .SS0 or .SS1 extension (predefined scenario/game)\n"
         "  -td         Encode or decode the TITLE.DAT file\n"
         "  -%c          %s\n"
         "  -h          More help\n"
         "\n",
         SAFE_CHARS_HIGH ? '7' : '8',  /* display the non-default choice */
         SAFE_CHARS_HIGH ?             /* same here */
                       "Restrict file names to a subset of US-ASCII" :
                       "Allow extended Latin-1 characters in file names");
  else fprintf(msgout,
         "List of all options\n"
         "====================\n"
         "Operation mode:\n"
         "  -c        Correct the title and file checksum of a savegame if necessary,\n"
         "            write back to the same file unless <output> is specified\n"
         "  -d        Assume that <input-file> is a savegame, decode to a 'big' file\n"
         "  -e        Assume that <input-file> is a 'big' file, encode as a savegame\n"
         "The default is to decode if <input-file> is recognized as a savegame,\n"
         "and to encode otherwise.\n"
         "\n"
         "Savegame naming options:\n"
         "  -trt, -trc, -trh, -try, -trp\n"
         "            Override the third character of the savegame file name.\n"
         "            Read the documentation for details.\n"
         "  -sv0, -sv1, -sv2, -ss0, -ss1\n"
         "            Specify the file extension (default is -sv1)\n"
         "\n"
         "'Big' file naming options:\n"
         "  -7        Use 7-bit characters (a subset of US-ASCII)%s\n"
         "  -8        Use 8-bit characters (a subset of Latin-1)%s\n"
         "  -xca      Escape unsafe characters using the \"@\" prefix (default)\n"
         "  -xcu, -u  Escape unsafe characters using the \"%%\" prefix (URL-style)\n"
         "  -xch      Escape unsafe characters using the \"#x\" prefix\n"
         "  -xcp      Escape unsafe characters using the \"_0x\" prefix\n"
         "If none of the -xc{a,u,h,p} options is specified, SV1Codec defaults to -u -xca.\n"
         "Read the documentation for more information.\n"
         "\n"
         "Savegame format options:\n"
         "  -ttd      Assume a Transport Tycoon Deluxe file\n"
         "  -tto      Assume a Transport Tycoon ('Original') file\n"
         "  -ttpd     Assume a Transport Tycoon Playable Demo file\n"
         "Note: The default is to detect the file type automatically.\n"
         "  -td       For encoding: encode as the appropriate TITLE.DAT file\n"
         "            For decoding or checksum correction: assume that <input-file>\n"
         "            is the TITLE.DAT file, even if its name is different\n"
         "  -title <title>, -name <title>\n"
         "            Use <title> as the savegame title\n"
         "  -m        Use maximum compression when encoding\n"
         "\n"
         "Ouptut control:\n"
         "  -stdmsg   Send normal messages to stdout, error messages to stderr\n"
         "  -stdout   Send all messages to stdout\n"
         "  -stderr   Send all messages to stderr\n"
         "\n"
         "Miscellaneous:\n"
         "  -f        Ignore non-fatal errors (force processing)\n"
         "  -h        This help message\n"
         "\n",
         SAFE_CHARS_HIGH ? "" : " (default)",
         SAFE_CHARS_HIGH ? " (default)" : "");

  fprintf(msgout, "Copyright (C) 2002 by Marcin Grzegorczyk.  See sv1codec.txt for more details.\n");
}



PUBLIC_DEFINITION
void UI_NoInputFile(void)
{
  UI_ErrorPrefix();
  fprintf(msgerr, "No input file specified\n");
}



PUBLIC_DEFINITION
void UI_GameRequiresMoreVehMsg(int ttdp)
/*
 Called after a savegame has been successfully encoded,
 if it uses the TTDPatch extended vehicle array feature 
*/
{
  fprintf(msgout, "Note: Game requires TTDPatch with 'morevehicles' option set to %d or more.\n", ttdp);
}



PUBLIC_DEFINITION
Bool UI_TitleChksumBad_AskFix(uint16_t ctc, uint16_t ftc)
/*
 Called in checksum-fix mode when the title checksum turns out to be wrong.
 'ctc' is the computed title checksum, 'ftc' is the title checksum stored
 in the file.

 [Reserved for an interactive version:]
 Return true if we're to fix the checksum and proceed, false to abort.
*/
{
  fprintf(msgout, "Title checksum is incorrect (computed: 0x%.4X, in file: 0x%.4X).\n",
                  (unsigned)ctc, (unsigned)ftc);
  return 1;
}



PUBLIC_DEFINITION
void UI_TitleChksumFixedMsg(void)
/*
 Called if the title checksum has been changed, and before recomputing
 the file checksum.
*/
{
  fprintf(msgout, "Title checksum fixed, recomputing file checksum...");
}



PUBLIC_DEFINITION void (*const UI_FileChksumRecomputedMsg)(void) = UI_OpDoneMsg;
/*
 Called after recomputing the file checksum.
*/



PUBLIC_DEFINITION
void UI_FileChksumFixedMsg(void)
/*
 Called if the file checksum has been fixed.
*/
{
  fprintf(msgout, "File checksum fixed.\n");
}



PUBLIC_DEFINITION
void UI_FileChksumOKNoActionMsg(void)
/*
 Called if both checksums are OK.
*/
{
  fprintf(msgout, "Both the title and the file checksums are correct. No action taken.\n");
}



PUBLIC_DEFINITION
void UI_TryNoArgsForHelp(void)
{
  fprintf(msgerr, "Try \"sv1codec\" (no arguments) for help.\n");
}

